
#include <cmath>
#include <iostream>
#include <cstdlib>

using namespace std;


#ifdef __APPLE__
  #include <GLUT/glut.h>         /* glut.h includes gl.h and glu.h*/
#else
  #include <GL/glut.h>         /* glut.h includes gl.h and glu.h*/
#endif


//*****************************************************************
// Area 0
// Begin: Define symbolic names for the action keys and 
//			the interface modes
//*****************************************************************
#define		X_up   'k'
#define		X_down   'j'
#define		Y_up   'i'
#define		Y_down   'm'
#define		Z_up   'l'
#define		Z_down   'h'
#define		KEY_TO_UNKNOWN_ACTION   0


#define TO_MOVE_THIS_SELECTED_OBJECT   1
#define TO_MOVE_THIS_SELECTED_CAMERA   2


int currentInterfaceInputMode = TO_MOVE_THIS_SELECTED_OBJECT;


//*****************************************************************
//*****************************************************************


//*****************************************************************
// Area 1
// Begin: Globa variables about the coordinates of the center of cube 
//			and the size of the cube;
//*****************************************************************
double cubeX=0, cubeY=0, cubeZ=0, cubeSize = 15;

//*****************************************************************
//*****************************************************************



//*****************************************************************
// Area 2
// Begin: Globa variables about the camera location and orientation
//*****************************************************************

//Camera on y = 0 plan, circling around (0,0,0)
//The following variable records the current angle between the vector (1,0,0) and (eyex, eyey, eyez)
double C_angle;

const double PI = 3.14156;

//Camera on y = 0 plan, circling around (0,0,0)
//The following variable records the radius of the camera's orbit
double C_Radius;

//Camera on y = 0 plan, circling around (0,0,0)
//The following variable records the rotation speed of the camera
double C_increment;

//Camera on y = 0 plan, circling around (0,0,0)
//Recording the currnet position of the camera.
double eyex, eyey, eyez; 

//Camera on y = 0 plan, circling around (0,0,0)
//Specifies the position of the point looked at
double centerx, centery, centerz; 
	
//Specifies the direction of the up vector. 
double upx, upy, upz;

//*****************************************************************
//*****************************************************************




//*****************************************************************
// Area 3
// Begin: THree simple utility functions
//*****************************************************************

/* return a random float in the range 0.0 ... 1.0 */
float randfloat(void)
{
	return (rand() % 1000)/1000.0;
}


void myCircle(float x0, float y0, float z0, float r, int numPoints)
{
	double x, y, z, thet;
	// inc is how many degrees separate each vertex; a circle made with 720 points
	// would thus have inc = 0.5 degrees
	double inc = 360.0/numPoints; 
	// c is used to convert from degrees to radians
	double c=3.14159/180.0;

	z=z0;

	glBegin(GL_LINE_LOOP);
	for (thet=-180.0; thet<180.0;thet+=inc) {
		x=r*sin(c*thet) + x0;
		y=r*cos(c*thet) + y0;
		glVertex3d(x,y,z);
	}
	glEnd();
}



void myFancyCircle(float x0, float y0, float z0, float r, int numPoints)
{
	double x, y, z, thet;
	double inc = 360.0/numPoints;
	double c=3.14159/180.0;

	z=0;

	
	glLineWidth(5.0);
	glBegin(GL_LINE_LOOP);
	for (thet=-180.0; thet<180.0;thet+=inc) {
		x=r*sin(c*thet) + x0;
		y=r*cos(c*thet) + y0;
		glColor3f(randfloat(),randfloat(),randfloat());
		glVertex3d(x,y,z);
	}
	glEnd();
	
	glLineWidth(1.0); //restore default line width
	glColor3f(1.0,1.0,1.0); //restore default color
}

//*****************************************************************
//*****************************************************************




//*****************************************************************
// Area 4
// Begin: Two initialization functions 
//	      the camaera movement function, and the display function
//*****************************************************************
void init()
{
	/* The scale variable is used for the first four parameters of glOrtho below */
	/* so that the viewing box can be used to view circles with radii greater */
	/* than 2.0. */
	float scale = 30.0;

	/* set clear color to black */
	glClearColor (0.0, 0.0, 0.0, 0.0);

	/* set fill  color to white */
	glColor3f(1.0, 1.0, 1.0);

	/* set up standard orthogonal view with clipping */
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	//glOrtho(-scale, scale, -scale, scale, 0.1, 5*scale);
	gluPerspective(90, 1, 0.1, 100);

}


void initCameraSetting()
{	
//The following variable records the current angle between the vector (1,0,0) and (eyex, eyey, eyez)
	C_angle = 0;

//Camera on y = 0 plan, circling around (0,0,0)
//The following variable records the radius of the camera's orbit
	C_Radius = 20;

//Camera on y = 0 plan, circling around (0,0,0)
//The following variable records the rotation speed of the camera
	C_increment = 2*PI / (360*2);

//Recording the currnet position of the camera.
	eyex = C_Radius; eyey = C_Radius; eyez = C_Radius;//C_Radius;

//Specifies the position of the point looked at as (0,0,0).
	centerx = 0; centery = 0; centerz=0;

//Specifies the direction of the up vector. 
	upx = 0; upy=1;  upz=0;

	//glMatrixMode(GL_MODELVIEW);
	//glLoadIdentity();
	//gluLookAt(eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz);

	//glMatrixMode (GL_PROJECTION);

}



void moveCamera()
{	
	//Camera on y = 0 plan, circling around (0,0,0)
	//The following variable records the current angle between the vector (1,0,0) and (eyex, eyey, eyez)
	C_angle += C_increment;


	if ( C_angle > 2*PI)
		C_angle -= 2*PI;

	eyex = C_Radius * cos(C_angle);
	eyey = C_Radius;
	eyez = C_Radius * sin(C_angle);


	glutPostRedisplay();

}


void myKeyboard(unsigned char key, int x, int y)
{   static bool isSpin = false;
	isSpin = !isSpin;

	if (isSpin)
		glutIdleFunc(moveCamera);
	else
		glutIdleFunc(NULL);

}

void drawAxesXYZ()
{	//Draw XYZ axes
	glBegin(GL_LINES);
	
		glColor3f(1,0,0);
		glVertex3f(0,0,0);
		glVertex3f(15,0,0);

		glColor3f(0,1,0);
		glVertex3f(0,0,0);
		glVertex3f(0,15,0);

		glColor3f(0,0,1);
		glVertex3f(0,0,0);
		glVertex3f(0,0,15);

	glEnd();

	//Show the X, Y, Z letters
	glColor3f(1,0,0);
	glRasterPos3f(5, 0, 0);
	glutBitmapCharacter(GLUT_BITMAP_9_BY_15, 'X');

	glColor3f(0, 1, 0);
	glRasterPos3f(0, 5, 0);
	glutBitmapCharacter(GLUT_BITMAP_9_BY_15, 'Y');

	glColor3f(0, 0, 1);
	glRasterPos3f(0, 0, 5);
	glutBitmapCharacter(GLUT_BITMAP_9_BY_15, 'Z');

}


void display(void)
{
	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();
	gluLookAt(eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz);

	float rvec[5] = {0.5, 0.75, 1, 2, 5};
	
	/* clear window */
	glClear(GL_COLOR_BUFFER_BIT);
	
	for (int i=0; i<5; i++) {	
		myCircle(0, 0, 0, rvec[i], 360);
	}
	
	drawAxesXYZ();

	myFancyCircle(1, 1, 0, 2.0, 60);
	myFancyCircle(-6, -4, 0, 3.0, 30);
	myFancyCircle(4, -3, 0, 1.0, 30);

	glPushMatrix();
	glTranslated(cubeX,cubeY, cubeZ);
	glutWireCube(cubeSize);
	glPopMatrix();

	
	/* swap and flush GL buffers */
	glutSwapBuffers();

}
//*****************************************************************
//*****************************************************************



//*****************************************************************
// Area 5
// Begin: all the mouse related functions
//*****************************************************************

//Prototype information
//*****************************************************
// Map a mouse motion to an action key
//*****************************************************
char mapMouseMotionToAction(double mouseMotionX, double  mouseMotionY);


//Prototype information
//*****************************************************
//Trigger an action according to what is the key and what is the current interface mode
//*****************************************************
void triggerActionAccordingToInterfaceMode(unsigned char key);


//Prototype information
//*******************************************************************
//Given  the world coordinates x,y,z of a point
//Determine the window coordinates winX, winY, winZ of the point
//    based on the settings of model view, projection, and viewport
//*******************************************************************
bool projectToScreenCoordinates
      (     double x,               double y,         double z,
            double& winX,           double& winY,     double& winZ
	);
      

//The following is the mouse motion function 
//*******************************************************************
//This function processes the mouse motion and act accordingly
//*******************************************************************

void myMouseMotion(int x, int y)

{ 		//Two static variables to remember the previous mouse location
		static int oldX = -1;
        static int oldY = -1;


        if (oldX == -1 && oldY == -1)

        {   oldX = x;  

            oldY = y;

            return;

        }


		//Determine the motion vector 
        double motionX=0, motionY=0;

            motionX =   x - oldX;

            motionY = -(y - oldY);



		//Disregard tiny motions
        static const double MOTION_LOWER_THRESHOLD = 5.0;

        if ( fabs(motionX) + fabs(motionY) < MOTION_LOWER_THRESHOLD )

            return;


 		//Disregard wild motions
       static const double MOTION_UPPER_THRESHOLD = 45.0;

        if ( fabs(motionX) + fabs(motionY) > MOTION_UPPER_THRESHOLD )

        {

            oldX = x;  

            oldY = y;

            return;

        }

 		//Remember this mouse location
        oldX = x;  

        oldY = y;

        char key;



 		//You can watch the console outputs to see what are happening
		cout << endl << endl << "**************************************************" << endl;
        cout << "mouse position, x=" << x <<" y=" << y << endl

                << "mouse motion, motionX=" << motionX <<" motionY=" << motionY << endl

                << " windowID=" << glutGetWindow() << endl;		//Interpreter the motion vector an action

		key = mapMouseMotionToAction(motionX, motionY);


 		//Trigger the action accordingly
       triggerActionAccordingToInterfaceMode(key);


		switch (key)
		{
			case X_up:
               cout << "Action= X_up" << endl << endl;
			   break;
			case X_down:
               cout << "Action= X_down" << endl << endl;
			   break;
			case Y_up:
               cout << "Action= Y_up" << endl << endl;
			   break;
			case Y_down:
               cout << "Action= Y_down" << endl << endl;
			   break;
			case Z_up:
               cout << "Action= Z_up" << endl << endl;
			   break;
			case Z_down:
               cout << "Action= Z_down" << endl << endl;
			   break;
			default:
               cout << "Action= UNKNOWN" << endl << endl;
		}
		cout << "**************************************************" << endl;


 
		//Force a redisplay of the scene
        glutPostRedisplay();

}



//*****************************************************
//Trigger an action according to what is the key and what is the current interface mode
//*****************************************************
void triggerActionAccordingToInterfaceMode(unsigned char key)
{	

	const double UNIT_MOVE_IN_XYZ = 0.1;

	switch ( currentInterfaceInputMode )
	{

		case TO_MOVE_THIS_SELECTED_OBJECT:
			if (key == X_up)
				// moveThisObjectAlongAxis(X,+UNIT_MOVE_IN_XYZ);
				cubeX += UNIT_MOVE_IN_XYZ;

			if (key == X_down)
				// moveThisObjectAlongAxis(X,-UNIT_MOVE_IN_XYZ);
				cubeX -= UNIT_MOVE_IN_XYZ;

			if (key == Y_up)
				// moveThisObjectAlongAxis(Y,UNIT_MOVE_IN_XYZ);
				cubeY += UNIT_MOVE_IN_XYZ;

			if (key == Y_down)
				// moveThisObjectAlongAxis(Y,-UNIT_MOVE_IN_XYZ);
				cubeY -= UNIT_MOVE_IN_XYZ;


			if (key == Z_up)
				// moveThisObjectAlongAxis(Z,UNIT_MOVE_IN_XYZ);
				cubeZ += UNIT_MOVE_IN_XYZ;

			if (key == Z_down)
				// pmoveThisObjectAlongAxis(Z,-UNIT_MOVE_IN_XYZ);
				cubeZ -= UNIT_MOVE_IN_XYZ;


			break;


	    //
		//	If you have other interface modes, you can add more cases below.
		//

		//case TO_MOVE_THIS_SELECTED_CAMERA: 
		//	if (key == X_up)
		//		// moveThisCameraAlongAxis(X,UNIT_MOVE_IN_XYZ);
		//	if (key == X_down)
		//		// moveThisCameraAlongAxis(X,-UNIT_MOVE_IN_XYZ);
		//	if (key == Y_up)
		//		// moveThisCameraAlongAxis(Y,UNIT_MOVE_IN_XYZ);
		//	if (key == Y_down)
		//		// moveThisCameraAlongAxis(Y,-UNIT_MOVE_IN_XYZ);
		//	if (key == Z_up)
		//		// pmoveThisCameraAlongAxis(Z,UNIT_MOVE_IN_XYZ);
		//	if (key == Z_down)
		//		// moveThisCameraAlongAxis(Z,-UNIT_MOVE_IN_XYZ);
		//	break;


		//case TO_MOVE_LOOKED_AT_POINT_OF_THIS_SELECTED_CAMERA: 
		//	if (key == X_up)
		//		// moveLookedAtPointOfThisCameraAlongAxis(X,UNIT_MOVE_IN_XYZ);
		//	if (key == X_down)
		//		// moveLookedAtPointOfThisCameraAlongAxis(X,-UNIT_MOVE_IN_XYZ);
		//	if (key == Y_up)
		//		// moveLookedAtPointOfThisCameraAlongAxis(Y,UNIT_MOVE_IN_XYZ);
		//	if (key == Y_down)
		//		// pmoveLookedAtPointOfThisCameraAlongAxis(Y,-UNIT_MOVE_IN_XYZ);
		//	if (key == Z_up)
		//		// moveLookedAtPointOfThisCameraAlongAxis(Z,UNIT_MOVE_IN_XYZ);
		//	if (key == Z_down)
		//		// moveLookedAtPointOfThisCameraAlongAxis(Z,-UNIT_MOVE_IN_XYZ);
		//	break;							


		default:
			break;
	}

}


//*****************************************************
// Map a mouse motion to an action key
//*****************************************************
char mapMouseMotionToAction(double mouseMotionX, double  mouseMotionY)
{

	//The following code segment determines the corresponding
	//    unit motion vectors aglong the X, Y, Z axes respectively, and
	//	  store the information in (axisX_diffX,	axisX_diffY)
	//    (axisY_diffX,	axisY_diffY), (axisZ_diffX,	axisZ_diffY) respectively.

	double axisX_newX,      axisX_newY,       axisX_newZ;

	double axisY_newX,      axisY_newY,       axisY_newZ;

	double axisZ_newX,      axisZ_newY,       axisZ_newZ;

	double origin_newX, origin_newY,    origin_newZ;



	projectToScreenCoordinates    (     1,    0,    0,   

										axisX_newX, axisX_newY, axisX_newZ

									);


	projectToScreenCoordinates    (     0,    1,    0,   

										axisY_newX, axisY_newY, axisY_newZ

								   );



	projectToScreenCoordinates    (     0,    0,    1,   

										axisZ_newX, axisZ_newY, axisZ_newZ

									);



	projectToScreenCoordinates    (     0,    0,    0,   

										origin_newX, origin_newY, origin_newZ

									);

	

	double axisX_diffX,     axisX_diffY;

	double axisY_diffX,     axisY_diffY;

	double axisZ_diffX,     axisZ_diffY;



	axisX_diffX = axisX_newX - origin_newX;

	axisX_diffY = axisX_newY - origin_newY;



	axisY_diffX = axisY_newX - origin_newX;

	axisY_diffY = axisY_newY - origin_newY;



	axisZ_diffX = axisZ_newX - origin_newX;

	axisZ_diffY = axisZ_newY - origin_newY;





	double      len_MouseMotion,

				len_ProjectedAxisX,

				len_ProjectedAxisY,

				len_ProjectedAxisZ;



	len_MouseMotion

		= sqrt( mouseMotionX *  mouseMotionX +  mouseMotionY *  mouseMotionY);



	len_ProjectedAxisX

		= sqrt(axisX_diffX * axisX_diffX + axisX_diffY * axisX_diffY);



	len_ProjectedAxisY

		= sqrt(axisY_diffX * axisY_diffX + axisY_diffY * axisY_diffY);



	len_ProjectedAxisZ

		= sqrt(axisZ_diffX * axisZ_diffX + axisZ_diffY * axisZ_diffY);



	static const double THRESHOLD_LENGTH = 0.001;

	if    (     len_MouseMotion               < THRESHOLD_LENGTH      ||

				len_ProjectedAxisX            < THRESHOLD_LENGTH      ||

				len_ProjectedAxisY            < THRESHOLD_LENGTH      ||

				len_ProjectedAxisZ            < THRESHOLD_LENGTH     

		)

		return KEY_TO_UNKNOWN_ACTION;





	mouseMotionX  /= len_MouseMotion;

	mouseMotionY  /= len_MouseMotion;



	axisX_diffX /= len_ProjectedAxisX;

	axisX_diffY /= len_ProjectedAxisX;



	axisY_diffX /= len_ProjectedAxisY;

	axisY_diffY /= len_ProjectedAxisY;



	axisZ_diffX /= len_ProjectedAxisZ;

	axisZ_diffY /= len_ProjectedAxisZ;



	//Determine the theta angles  thetaX, thetaY, thetaZ between the mouse motion vector
	// and (axisX_diffX, axisX_diffY), and (axisY_diffX, axisY_diffY), and
	// (axisZ_diffX, axisZ_diffY) respectively.

	double thetaX, thetaY, thetaZ;

	thetaX = acos( axisX_diffX * mouseMotionX + axisX_diffY * mouseMotionY);

	thetaY = acos( axisY_diffX * mouseMotionX + axisY_diffY * mouseMotionY);

	thetaZ = acos( axisZ_diffX * mouseMotionX + axisZ_diffY * mouseMotionY);



	cout << endl << endl

			<< "X axis unit vector=(" << axisX_diffX << ", " << axisX_diffY << ") " << endl

			<< "Y axis unit vector=(" << axisY_diffX << ", " << axisY_diffY << ") " << endl

			<< "Z axis unit vector=(" << axisZ_diffX << ", " << axisZ_diffY << ") " << endl;



	cout    << " Degree between X axis and mouse motion=" << thetaX*180.0/3.14159 << endl

			<< " Degree between Y axis and mouse motion=" << thetaY*180.0/3.14159 << endl

			<< " Degree between Z axis and mouse motion=" << thetaZ*180.0/3.14159 << endl;



	static const double THRESHOLD_RADIAN = 0.5;

	static const double PI = 3.14159;


	//Determine the action based on the angles of thetaX, thetaY, thetaZ 

	if (  thetaX                  <     THRESHOLD_RADIAN  )

		return X_up;

	if (  ( PI - thetaX )   <     THRESHOLD_RADIAN  )

		return X_down;



	if (  thetaY                  <     THRESHOLD_RADIAN  )

		return Y_up;

	if (  ( PI - thetaY )   <     THRESHOLD_RADIAN  )

		return Y_down;



	if (  thetaZ                  <     THRESHOLD_RADIAN  )

		return Z_up;

	if (  ( PI - thetaZ )   <     THRESHOLD_RADIAN  )

		return Z_down;



}


 

//*******************************************************************
//Given  the world coordinates x,y,z of a point
//Determine the window coordinates winX, winY, winZ of the point
//    based on the settings of model view, projection, and viewport
//*******************************************************************

bool projectToScreenCoordinates

      (     double x,               double y,         double z,

            double& winX,           double& winY,     double& winZ

      )

{ 

      int windowWidth, windowHeight;

	  windowWidth = glutGet(GLUT_WINDOW_WIDTH);
	  windowHeight = glutGet(GLUT_WINDOW_HEIGHT);

      if (windowWidth <= 0 || windowHeight <= 0)

            return false;

 

      //Reset the modelview matrix and store the information in modelMatrix

	  GLdouble modelMatrix[16];

      glMatrixMode(GL_MODELVIEW);

      glLoadIdentity();

	  gluLookAt(eyex, eyey, eyez, centerx, centery, centerz, upx, upy, upz);

      glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);



      //Reset the projection matrix and store the information in projMatrix

      GLdouble projMatrix[16];

      glMatrixMode(GL_PROJECTION);

	  glLoadIdentity ();

	  gluPerspective(90, 1, 0.1, 100);
	  
      glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);

      
	  //store the information of the view port into viewport
      GLint viewport[4];
      glGetIntegerv(GL_VIEWPORT, viewport);

 
	  //Determine the windo coordinates (winX, winY, winZ) of the point (x,y,z)
      bool flag =

            gluProject  (     x, y, z,

                              modelMatrix, projMatrix, viewport,

                              & winX, &winY, & winZ

                              ) ;

 

      return flag;

}

 
//*****************************************************************
// Area 6
// Begin: the main function
//*****************************************************************

int main(int argc, char** argv)
{
	/* Initialize mode and open a window in upper left corner of screen */

	glutInit(&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB);
	glutInitWindowSize (250, 250); 
	glutInitWindowPosition (0, 0);

	glutCreateWindow("Mouse-motion-event handler");
	glutDisplayFunc(display);
	glutMotionFunc(myMouseMotion);
	glutKeyboardFunc(myKeyboard);
	glutIdleFunc(NULL);

	init();
	initCameraSetting();

	glutMainLoop();

}
//*****************************************************************
//*****************************************************************

